/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/

//jule is the number of propagators
#include "yaml-cpp/yaml.h"
#include "sort_rules.h"
#include <fstream>
#include "kira.h"
#include "integral.h"
#include "ReadYamlFiles.h"
#include "tools.h"
#include "sqlite3/sqlite3.h"
#include "dataBase.h"
#include <sys/stat.h>
#include "createSeeds.h"
#include "interface.h"
#include <omp.h>
#include <pthread.h>

#include <gzstream.h>

#include "integrals.h"

#include <algorithm>

using namespace pyred;
  
using namespace std;
using namespace GiNaC;
using namespace YAML;

static pthread_mutex_t mPreserveFermat;
static pthread_cond_t cPreserveFermat;

static Loginfo & logger = Loginfo::instance();

Kira::Kira (string name, int coreNumber_,int algebra_, string pyred_config_) {
  
  coreNumber = coreNumber_;
  algebra = algebra_;
  jobName = name;
  pyred_config = pyred_config_;
};

int Kira::look_up_symmetry(vector<BaseEquation*>& setUpEq, IBPIntegral *&integral, SYM *&sym,int topologyNUM){
  
  ex coef1, coef2, number;
  IBPVG Identitie_Integral, Identitie_Integral2;
  int sec=integral->characteristics[SECTOR];
  
  for(vector<symmetries>::iterator symHandler = sym[sec].begin(); symHandler!=sym[sec].end(); symHandler++){

    if (symHandler->det == 0)
      continue;
    
    if (symHandler->symDOTS == 1 && (integral->characteristics[NUM]>0 /*|| (integral->characteristics[DOTS]-integral->characteristics[DENCOUNT])>1*/)){
      continue;
    }
    
//     if (symHandler->symDOTS == 0 && (integral->characteristics[DOTS]-integral->characteristics[DENCOUNT])>0)
//       continue;
    
//     cout<<symHandler->symDOTS<<" "<<(integral->characteristics[DOTS]-integral->characteristics[DENCOUNT])<< " " << integral->characteristics[NUM] << "\n";
    
    string topoName = collectReductions[ symHandler->topology ];
    string topoNameOriginal = collectReductions[topologyNUM];
    
    vector<vector<pair<ex,int> > > exTerm;
    ex ausd2;
    
    for (int j=0; j<topology[topoNameOriginal].jule;j++){
      if (symHandler->ing[j]<0){
        vector<pair<ex,int> > uj;
        if (integral->indices[j]<0){
          ausd2 = topology[topoNameOriginal].props[j].expand().subs(symHandler->subst,
                            subs_options::algebraic).expand();
          fs<lst>(ausd2,kinematic);
          ausd2 = ausd2.subs(topology[topoName].scal2Props[symHandler->externalSymmetry], subs_options::algebraic);
          ex traley;
          if((traley=ausd2.subs(suby,subs_options::algebraic))!=0){
            uj.push_back(pair<ex, int > (traley,-1));
          }
          for(int jopp=0; jopp<topology[topoNameOriginal].jule;jopp++){
            uj.push_back(pair<ex, int > (ausd2.coeff(topology[topoNameOriginal].propSymb[jopp],1),jopp));
          }
        }
        for (int iA=0; iA < -integral->indices[j];iA++){
          exTerm.push_back(uj);
        }
      }
    }
    
//     for (int j=0; j<topology[topoNameOriginal].jule;j++){
//       vector<pair<ex,int> > uj;
// //       if (integral->indices[j]<0){
//         ausd2 = topology[topoNameOriginal].props[j].expand().subs(symHandler->subst,
//                           subs_options::algebraic).expand();
//         fs<lst>(ausd2,kinematic);
//         ausd2 = ausd2.subs(topology[topoName].scal2Props[symHandler->externalSymmetry], subs_options::algebraic);
//         ex traley;
//         if((traley=ausd2.subs(suby,subs_options::algebraic))!=0){
//           uj.push_back(pair<ex, int > (traley,-1));
//         }
//         for(int jopp=0; jopp<topology[topoNameOriginal].jule;jopp++){
//           uj.push_back(pair<ex, int > (ausd2.coeff(topology[topoNameOriginal].propSymb[jopp],1),jopp));
//         }
// //       }
// //       for (int iA=0; iA < -integral->indices[j];iA++){
//         exTerm.push_back(uj);
// //       }
//     }
    
    
    
    IBPIntegral *newIntegral;  
    newIntegral = new IBPIntegral(topology[topoNameOriginal].jule);
    
    for (int j=0; j < topology[topoNameOriginal].jule;j++){
      newIntegral->indices[j] = 0;
    }
    for (int j=0; j<topology[topoNameOriginal].jule;j++){
      if (symHandler->ing[j] >= 0){
        newIntegral->indices[symHandler->ing[j]] = integral->indices[j];
      }
    }
    newIntegral->coefficient=1;
    
    newIntegral->characteristics[TOPOLOGY] = symHandler->topology;
    newIntegral->characteristics[SECTOR] = symHandler->sector;
    
    Identitie_Integral.push_back(newIntegral);
      
    for(vector<vector<pair<ex,int> > >::iterator kii=exTerm.begin();
        kii!=exTerm.end();kii++){
      for(ItIBPVG i = Identitie_Integral.begin(),
            end = Identitie_Integral.end(); i != end ; ++i){
        for(vector<pair<ex,int> >::iterator lii=(*kii).begin();
            lii!=(*kii).end();lii++){

          newIntegral=new IBPIntegral(topology[topoNameOriginal].jule);
          
          for ( int j = 0; j<topology[topoName].jule; j++) {
            newIntegral->indices[j] = (*i)->indices[j];
          }
          if((*lii).first!=0){
            if((*lii).second!=-1)
              --newIntegral->indices[(*lii).second];
            newIntegral->coefficient = 
              (((*lii).first)*(*i)->coefficient).expand();
            newIntegral->generate_characteristics(symHandler->topology);
            Identitie_Integral2.push_back(newIntegral);
          }
          else
            delete newIntegral;
        }
        delete (*i);
      }
      Identitie_Integral.clear();
      Identitie_Integral=Identitie_Integral2;
      Identitie_Integral2.clear();
    }
    newIntegral = new IBPIntegral(topology[topoNameOriginal].jule);
    for ( int j = 0; j<topology[topoNameOriginal].jule; j++) {
      newIntegral->indices[j] = integral->indices[j];
    }
    newIntegral->coefficient=-1;
    newIntegral->characteristics[TOPOLOGY] = topologyNUM;
    newIntegral->characteristics[SECTOR] = sec;
    Identitie_Integral.push_back(newIntegral);

    BaseEquation * tokenEq = new BaseEquation(Identitie_Integral,topology[topoNameOriginal].jule,(int_seeds->integralMap), mastersMap);
    
    if((tokenEq->sort())>1){
      setUpEq.push_back(tokenEq);
    } 
    else {
      delete tokenEq;
    }
    
    for ( ItIBPVG i = Identitie_Integral.begin(); i != Identitie_Integral.end(); i++) {
      delete *i;
    }
    Identitie_Integral.clear();
  }
  return 1;
};

int Kira::setup_sym_eq(vector<BaseEquation*>& setUpEq, string name, int topologyNUM, int secNUM){
  
  if(!(int_seeds->aIBPint[topologyNUM][secNUM].size())){
    return 0;
  }
  
  logger << "Symmetry seeds: "<< int_seeds->aIBPint[topologyNUM][secNUM].size() << ";";
  
  for  ( auto it = int_seeds->aIBPint[topologyNUM][secNUM].begin(); it != int_seeds->aIBPint[topologyNUM][secNUM].end(); ++it) {
    
    IBPIntegral *integral = new IBPIntegral((*it),integralfamily.jule);
    
    int sec=integral->characteristics[SECTOR];
    
    if(topology[collectReductions[topologyNUM]].symVec[sec].size() && (integral->characteristics[NUM]<3)){//<3
      look_up_symmetry(setUpEq, integral,topology[collectReductions[topologyNUM]].symVec,topologyNUM);
    }
    
    if( topology[collectReductions[topologyNUM]].relVec[sec].size() && (integral->characteristics[NUM]<3) ){ // <3
      look_up_symmetry(setUpEq, integral,topology[collectReductions[topologyNUM]].relVec,topologyNUM);
    }

    delete integral;        
  }
  
  return 1;
  
};

int Kira::setup_ibp_li_eq(vector<BaseEquation*>& setUpEq, VE &identities, vector<std::uint64_t>& weights, string name, int topology){
  
  size_t nID=weights.size();
  if(!(identities.size())||!nID){
    return 0;
  }
  
  omp_set_dynamic(0);
  int nThreads = coreNumber;
  omp_set_num_threads(nThreads);
  
  omp_lock_t lck;
  omp_init_lock(&lck);
  
  #pragma omp parallel for schedule(dynamic) shared(lck)    
  for  (unsigned it = 0; it < nID; it++) {
    
    for(unsigned idenIt = 0; idenIt < identities.size(); idenIt++){
      
      int id = -1;
      
      pthread_mutex_lock(&mPreserveFermat);
	while(idFermats.empty())
	  pthread_cond_wait(&cPreserveFermat, &mPreserveFermat);
	id = idFermats.top();
	idFermats.pop();
      pthread_mutex_unlock(&mPreserveFermat);
	
      BaseEquation* tokenEq = new BaseEquation(*identities[idenIt],weights[it],(int_seeds->integralMap),topology, fermat, id, mastersMap);
      
      pthread_mutex_lock(&mPreserveFermat);
        idFermats.push(id);
      pthread_cond_broadcast(&cPreserveFermat);
      pthread_mutex_unlock(&mPreserveFermat);
      
      if(tokenEq->sort()){
      
        omp_set_lock(&lck);

          setUpEq.push_back(tokenEq);

        omp_unset_lock(&lck);        
      }
      else{

        delete tokenEq;
      }
    }
  }



  omp_destroy_lock(&lck);
  
  return 1;
  
};

void Kira::initiate_system_of_equations(int nOfTopology){
  
  Clock clock;
  
  mkdir((outputDir+"/tmp/"+integralfamily.name).c_str(), S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
  
  logger << "***********************************************************\n";
  logger << "Generate equations\n";
  logger << "***********************************************************\n";
  
  ofstream outputConfig((outputDir+"/tmp/"+integralfamily.name+"/SYSTEMconfig.tmp").c_str(), std::fstream::app);
  outputConfig.close();
  
  for(int i = 0; i < nOfTopology; i++) {
    
    for(int j = 0; j < int_seeds->nSector[i]; j++){
      
      vector<BaseEquation*> setUpEq;
      
      int secNUM = int_seeds->mappingSectorReverse[i][j];
      
      if(file_exists((outputDir+"/tmp/"+integralfamily.name+"/SYSTEM"+"sym"+"_"+
      collectReductions[i]+"_"+to_string(secNUM)+".gz").c_str()))
	continue;
      
	
      if(!(topology[collectReductions[i]].identitiesIBP.size())||!(int_seeds->aIBPint[i][j]).size() ){
        continue;
      }
      
      logger << "Topology: " << collectReductions[i] << "; ";
      
      logger << "Sector: " << secNUM << "; ";
      
      setup_sym_eq(setUpEq,"SYM",i,j); 
      
      logger <<"\n";
      
      sort(setUpEq.begin(), setUpEq.end(), sort_rules4);
      
      write_init_system(setUpEq, collectReductions[i],secNUM,"sym");
      
      ifstream inputConfig;
      unsigned tmpEqNumber = 0;
      inputConfig.open ((outputDir+"/tmp/"+integralfamily.name+"/SYSTEMconfig").c_str());
      inputConfig >> tmpEqNumber;
      tmpEqNumber += setUpEq.size();
      inputConfig.close();
      
      ofstream outputConfig;
      outputConfig.open ((outputDir+"/tmp/"+integralfamily.name+"/SYSTEMconfig").c_str());
      outputConfig << tmpEqNumber << endl;
      outputConfig.close();
      
      
      rename( (outputDir+"/tmp/"+integralfamily.name+"/SYSTEM"+"sym"+"_"+collectReductions[i]
      +"_"+to_string(secNUM)+".tmp.gz").c_str(), (outputDir+"/tmp/"+integralfamily.name
      +"/SYSTEM"+"sym"+"_"+collectReductions[i]+"_"+to_string(secNUM)+".gz").c_str());
      
      
      for ( unsigned it = 0; it < setUpEq.size(); it++) {
	  delete setUpEq[it];
      }
      setUpEq.clear();
    }
  }

  for(int i = 0; i < nOfTopology; i++) {
    
    omp_set_nested(1);
    omp_set_dynamic(0);
    int nThreads = coreNumber;
    if(maxParallelSectors)
      omp_set_num_threads(maxParallelSectors);
    else
      omp_set_num_threads(nThreads);
    
    #pragma omp parallel for schedule(dynamic)
    for(int j = 0; j < int_seeds->nSector[i]; j++){
      
      Clock clock2;
      
      vector<BaseEquation*> setUpEq;
      
      int secNUM = int_seeds->mappingSectorReverse[i][j];
      
      if(file_exists((outputDir+"/tmp/"+integralfamily.name+"/SYSTEM"+"_"+
      collectReductions[i]+"_"+to_string(secNUM)+".gz").c_str()))
	continue;
	
      if(!(topology[collectReductions[i]].identitiesIBP.size())||!(int_seeds->aIBPint[i][j]).size() ){
        continue;
      }
      #pragma omp critical
      {
        logger << "Topology: " << collectReductions[i] << "; ";
      
        logger << "Sector: " << secNUM << "; ";
        logger << " seeds: " << int_seeds->aIBPint[i][j].size() << "; \n";
      }
      setup_ibp_li_eq(setUpEq,topology[collectReductions[i]].identitiesIBP, int_seeds->aIBPint[i][j], "IBP", i);
      
      
      
      setup_ibp_li_eq(setUpEq,topology[collectReductions[i]].identitiesLI, int_seeds->aIBPint[i][j], "LI", i);
      
      
      sort(setUpEq.begin(), setUpEq.end(), sort_rules4);
      
      write_init_system(setUpEq, collectReductions[i], secNUM, "");
      
      #pragma omp critical
      {
	ifstream inputConfig;
	unsigned tmpEqNumber = 0;
	inputConfig.open ((outputDir+"/tmp/"+integralfamily.name+"/SYSTEMconfig").c_str());
	inputConfig >> tmpEqNumber;
	tmpEqNumber += setUpEq.size();
	inputConfig.close();
	
	ofstream outputConfig;
	outputConfig.open ((outputDir+"/tmp/"+integralfamily.name+"/SYSTEMconfig").c_str());
	outputConfig << tmpEqNumber << endl;
	outputConfig.close();
	
	
	rename( (outputDir+"/tmp/"+integralfamily.name+"/SYSTEM"+"_"+collectReductions[i]
	+"_"+to_string(secNUM)+".tmp.gz").c_str(), (outputDir+"/tmp/"+integralfamily.name
	+"/SYSTEM"+"_"+collectReductions[i]+"_"+to_string(secNUM)+".gz").c_str());
	
        logger << "Topology: " << collectReductions[i] << "; ";
        logger << "Sector: " << secNUM << "; ";
        logger << "( " << clock2.eval_time() << " s )\n";
      }
      
      for ( unsigned it = 0; it < setUpEq.size(); it++) {
	  delete setUpEq[it];
      }
      setUpEq.clear();
    }
  }
  
// 	outputConfig << setUpEq.size() <<"\n";
	
	
// 	outputConfig.close();
  
  logger << "\n( " << clock.eval_time() << " s )\n";
};


vector<vector<int > > Kira::read_id2int(){
  
  ifstream fileInput;
  fileInput.open((outputDir+"/results/"+integralfamily.name+"/id2int").c_str());
  
  int jule = integralfamily.jule;
  string line;
  vector<vector<int> > arraySeed;
  
  while(getline(fileInput, line)) {
    
    //get string indices
    if(line.size()<7)
      continue;
    
    size_t found;
    
    if((found = line.find_first_of("#")) != string::npos){
      logger << "this line will be skept: " << line<< "\n";
      continue;
    }
    
    if((found = line.find_first_of("[")) == string::npos){
      logger << "this line will be skept: " << line<< "\n";
      continue;
    }
    
    line.erase(line.begin(),line.begin()+found+1);
    line.erase(remove_if(line.begin(), line.end(), ::isspace), line.end());
    
    if((found = line.find_first_of("]")) == string::npos){
      logger << "this line will be skept: " << line<< "\n";
      continue;
    }
    
    line.erase(line.begin()+found);//indices
    
    //convert string indices to integer array
    int indicesCounter;
    vector<int> inputSeed;
    int countN=0;
    while((found = line.find_first_of(",")) != string::npos){

      if((jule + 1)!=countN && (jule + 2)!=countN){
	countN++;
	line = line.substr(found+1);
	continue;
      }
      
      istringstream(line.substr(0,found))>>indicesCounter;
      inputSeed.push_back(indicesCounter);
      line=line.substr(found+1);
      
      countN++;
    }
    
    istringstream(line) >> indicesCounter;
    inputSeed.push_back(indicesCounter);
    arraySeed.push_back(inputSeed);
  }
  
  return arraySeed;
}

void Kira::run_pyred(std::vector<std::uint64_t>& mandatory, 
                     std::vector<std::uint64_t>& optional){
  
  
  
  Clock clock;
  logger << "\n***** Starting PYRED **************************************\n";
  
  load_pyred_on_the_fly(mandatory,optional);
  
  logger << "( " << clock.eval_time() << " s )\n";
};

int Kira::complete_reduction(){
  
  logger << "\n***** Run the back substitution ***********************\n";
  
  DataBase database(outputDir+"/results/kira.db");
  database.create_equation_table();
//   database.create_skipid_table();
  
//   if(integralfamily.topology > 0) {
//     string attachName = (outputDir+"/results/"+collectReductions[ (integralfamily.topology - 1) ] +"/kira.db");
//     logger << "Read results from: " << attachName<<".\n";
//     database.attach_table( attachName );
//   }
  database.DataBase::~DataBase();
  
  string inputName = outputDir+"/tmp/"+integralfamily.name+"/masters";
  ifstream input;
  
  vector<std::uint64_t> masterVector;
  
  if(file_exists(inputName.c_str())){  
    input.open(inputName.c_str());
    while(1){
      std::uint64_t ID;
      if(!(input >>ID)) break;
      
      auto mapContent = mastersMap.find(ID);
                    
      if(mapContent != mastersMap.end()){
        
        ID = mapContent->second;
      }
      masterVector.push_back(ID);
    }
    input.close();
  }
  
  auto ret = selectMastersReduction.equal_range(integralfamily.name);
  
  if (ret.first != selectMastersReduction.end()){
  
    for (auto itRet=ret.first; itRet!=ret.second; ++itRet){
          
      for(unsigned i = 0; i < masterVector.size(); i++){
          
        auto findMaster = find( (itRet->second).begin(),(itRet->second).end(), i+1 );
        
        if( findMaster == (itRet->second).end() ){
          
          masterVectorSkip.push_back( masterVector[i] );
        }
      }
      
      if( !load_back_substitution() )
	return 0;
      
      back_subs();
      clean_back_subs();
    }
  }
  else{
    
    if( !load_back_substitution() )
	return 0;
      
    back_subs();
    clean_back_subs();
  }
  return 1;
}

void Kira::clean_back_subs(){
  
  for ( unsigned it = 0, end = reihen+1; it != end; it++) {
    if (length[it]) {
      delete [] allEq[it][0];
    }
    if(allEq[it]!=0)
      delete [] allEq[it];
  }
  
  delete [] rdy2P;
  delete [] last_reduce;
  delete [] length;
  delete [] allEq;
  reduct2StartHere.clear();
  occurrence.clear();
  reverseLastReduce.clear();
  masterVectorSkip.clear();
}


void Kira::print_equation2(BaseIntegral*& integral,ogzstream& output){
  
  output << "Eq" << "\n";
  output << integral[0].id << "\n";
  output << integral[0].length << "\n";
  
  for (unsigned j=0; j<integral[0].length;j++){
    integral[j].length=integral[0].length;
    output << integral[j];
  }
}


void Kira::print_equation(BaseIntegral*& integral,ogzstream& output,ofstream& outputX,tuple<int,int/*,unsigned,int*/>& printControl,tuple<string,string>& dest){
  
  if (integral[0].characteristics[SECTOR] != get<0>(printControl) ||integral[0].characteristics[TOPOLOGY] != get<1>(printControl)){
    
    get<0>(printControl) = integral[0].characteristics[SECTOR];
    get<1>(printControl) = integral[0].characteristics[TOPOLOGY];
    
    
//     if(get<0>(printControl) != static_cast<int>(get<2>(printControl)) || get<1>(printControl) != get<3>(printControl)){
//       output.close();
//       outputX.close();
//       get<0>(dest) = integralfamily.name+"/SYSTEMx";
//       get<1>(dest) = "";
//       string name = outputDir+"/tmp/" + get<0>(dest) + "_" + collectReductions[get<1>(printControl)] + "_" + something_string(get<0>(printControl))+get<1>(dest);
//       outputX.open (name.c_str(),std::fstream::app);
//     }
//     else{
//       outputX.close();
//       output.close();    
//       get<0>(dest) = integralfamily.name+"/VER";
//       get<1>(dest) = ".tmp";
//       string name = outputDir+"/tmp/" + get<0>(dest) + "_" + collectReductions[get<1>(printControl)] + "_" + something_string(get<0>(printControl))+get<1>(dest)+".gz";
//       output.open (name.c_str());
//     }
//     cout << "koko: " << name << endl;
    
    output.close();    
    get<0>(dest) = integralfamily.name+"/VER";
    get<1>(dest) = "";
    string name = outputDir+"/tmp/" + get<0>(dest) + "_" + collectReductions[get<1>(printControl)] + "_" + something_string(get<0>(printControl))+get<1>(dest)+".gz";
    output.open (name.c_str());
  }
  
//   cout << "hoho: " <<(outputDir+"/tmp/" + get<0>(dest) + "_" + collectReductions[get<1>(printControl)] + "_" + something_string(get<0>(printControl))+".gz"+get<1>(dest)) << endl;
  
//   if(get<1>(dest) != ".tmp"){
// #pragma omp critical
//     {
// //       integral[0].coefficientString = "-1";
//       outputX << "Eq" << "\n";
//       outputX << integral[0].id << "\n";
//       outputX << integral[0].length << "\n";
//       
//       for (unsigned j = 0; j < integral[0].length; j++){
// 	
// 	integral[j].length = integral[0].length;
// 	outputX << integral[j];
//       }
//     }
//   }
//   else{

      output << "Eq" << "\n";
      output << integral[0].id << "\n";
      output << integral[0].length << "\n";
      
      for (unsigned j = 0; j < integral[0].length; j++){
	
	integral[j].length = integral[0].length;
	output << integral[j];
      }
//   }
}

void Kira::write_init_system(std::vector<BaseEquation*>& setUpEq, const std::string & topology, int secNUM, const std::string & add){
  
  ogzstream output;
  
  output.open ((outputDir+"/tmp/"+integralfamily.name+"/SYSTEM"+add+"_"+topology+"_"+to_string(secNUM)+".tmp.gz").c_str());
  

  for( size_t it = 0; it < setUpEq.size(); it++){
    print_equation2(setUpEq[it]->equation,output);
  }
  
  output.close();
};

void Kira::write_pyred(vector<vector<pair<std::uint64_t,string> > > & eqs,/*
  vector<vector<int> > & arraySeed,*/ vector<pyred::intid> & independent_eqnums){
  
  logger << "\nWrite with PYRED reduced system to hard disk \n";
  logger << "(--> "<<outputDir<<"/tmp/NUMconfig, "<<outputDir<<"/tmp/NUM###) \n";
  
  ofstream output;
  
  output.open ((outputDir+"/tmp/"+integralfamily.name+"/NUMconfig").c_str());
  output << eqs[eqs.size()-1][0].first << "\n";
  
  pair<unsigned,unsigned> printControl;
  printControl.first = -1;
  printControl.second = -1;
  
  for(unsigned it = 0; it < eqs.size(); it++){
    std::uint64_t id = eqs[it][0].first;
    
    auto iglback = pyred::Integral(id);
    auto property = iglback.properties(id);
    
    if (property.sector != printControl.first //sector
    || property.topology != printControl.second){ //topology
      printControl.first = property.sector; //sector
      printControl.second = property.topology; //topology
      output.close();
      string name = outputDir+"/tmp/" + integralfamily.name + "/NUM" + "_" + collectReductions[printControl.second] + "_" + something_string(printControl.first);
      output.open (name.c_str());
    }
    output << "Eq" << "\n";
    output << eqs[it][0].first << "\n";
    output << eqs[it].size() << "\n";
  
    for (size_t it2 = 0; it2 < eqs[it].size(); it2++){
      output << eqs[it].size() << " ";
      output << eqs[it][it2].second << " ";
      output << eqs[it][it2].first << " ";
      output << property.sector << " ";
      output << property.topology << " ";
      output << "0" << "\n"; //flag2
    }
  }
  
  output.close();
};

void Kira::write_triangular(/*unsigned sectorNumber, int k,*/ std::map<std::uint64_t, std::tuple<BaseIntegral ***, int> >& forwardRed, int flagy ){
  
  logger << "\nWrite the triangular system to "<<outputDir<<"/tmp/ on hard disk\n";
  logger << "(--> "<<outputDir<<"/tmp/"<<integralfamily.name<<"/VERconfig, ";
  logger <<outputDir<<"/tmp/"<<integralfamily.name<<"/VER###) \n";
  
  ogzstream output;
  ofstream outputX;
  
  tuple<string,string> dest;
  
  tuple<int,int/*,unsigned,int*/> printControl;
  get<0>(printControl) = -1;
  get<1>(printControl) = -1;
//   get<2>(printControl) = sectorNumber;
//   get<3>(printControl) = k;
  int rememberID=0;
  
  for( auto toIt = forwardRed.begin(); toIt != forwardRed.end(); toIt++){ //collect all keys to iterate
    if(get<1>(toIt->second)>0 && get<0>(toIt->second)[0][0][0].flag2 == 0){
      print_equation(get<0>(toIt->second)[0][0],output,outputX,printControl,dest);
      rememberID++;
    }
  }
  output.close();
  
// #pragma omp critical
//   {
//     ifstream inputConfig;
//     unsigned tempNumber = 0;
//     inputConfig.open ((outputDir+"/tmp/"+integralfamily.name+"/VERconfig").c_str());
//     
//     inputConfig >> tempNumber;
//     inputConfig.close();
//     
//     ofstream outputConfig;
//     outputConfig.open ((outputDir+"/tmp/"+integralfamily.name+"/VERconfig").c_str());
    
//     if(flagy == 1)
//       tempNumber += rememberID;
    
//     outputConfig << tempNumber << "\n";
//     outputConfig.close();
//   }
    
    ofstream outputConfig;
    outputConfig.open ((outputDir+"/tmp/"+integralfamily.name+"/VERconfig").c_str());
    outputConfig << rememberID << "\n";
    outputConfig.close();
  
//   rename( (outputDir+"/tmp/"+integralfamily.name+"/VER"+"_"+collectReductions[k]
//   +"_"+to_string(sectorNumber)+".tmp.gz").c_str(), (outputDir+"/tmp/"+integralfamily.name
//   +"/VER"+"_"+collectReductions[k]+"_"+to_string(sectorNumber)+".tmp2.gz").c_str());
//   
//   if(!flagy){
//     remove((outputDir+"/tmp/"+integralfamily.name
//       +"/SYSTEMx"+"_"+collectReductions[k]+"_"+to_string(sectorNumber)).c_str());
//   }
};

template<typename T>
void Kira::read_integral2(T& input, int eqLength,
			  std::map<std::uint64_t, std::tuple<BaseIntegral ***, int> >& forwardRed,
			  std::unordered_map<std::uint64_t,int>& rdy2F){
    
  BaseIntegral *integral = new BaseIntegral [eqLength];
  
  for (int j=0; j<eqLength;j++){
    
    if(!(input>>integral[j])) break;

    integral[j].length=eqLength;
    
//     auto mapContent = mastersReMap.find(integral[j].id);
//     
//     if(mapContent != mastersReMap.end()){
//       
//       integral[j].id = mapContent->second;
//     }
  }
  
  FWREDI mapContent = forwardRed.find(integral[0].id);
  
  if ( mapContent != forwardRed.end() ) {
    
    if(get<1>(mapContent->second) >= RED){
      
      BaseIntegral*** put2map = new BaseIntegral** [1];
      put2map[0] = new BaseIntegral * [get<1>(mapContent->second)+1];
              
      for(int kIt=0; kIt < get<1>(mapContent->second); kIt++){
        
        put2map[0][kIt] = get<0>(mapContent->second)[0][kIt];
      }
      
      delete [] get<0>(mapContent->second)[0];
      
      get<0>(mapContent->second)[0] = put2map[0];
    }
    
    get<0>(mapContent->second)[0][get<1>(mapContent->second)]=integral;
    get<1>(mapContent->second)++; 
  }
  else{
    
    BaseIntegral *** put2map = new BaseIntegral ** [1];
    put2map[0] = new BaseIntegral * [RED];
    put2map[0][0]=integral;
    forwardRed.insert(pair<std::uint64_t, tuple<BaseIntegral***, int> > (integral[0].id, make_tuple(put2map,1)));
    rdy2F.insert(pair<std::uint64_t, int> (integral[0].id, 0));
  }
}

void Kira::skip_integral(igzstream& input,int eqLength){
    
  BaseIntegral *integral = new BaseIntegral [eqLength];
  
  for (int j=0; j<eqLength;j++){
    
    if(!(input>>integral[j])) break;
  }
  delete [] integral;
}

void Kira::read_integral(BaseIntegral*& integral, igzstream& input, int eqLength){
    
  integral = new BaseIntegral [eqLength];
  
  for (int j = 0; j < eqLength; j++){
    
    if(!(input>>integral[j])) break;

    integral[j].length = eqLength;
    
//     auto mapContent = mastersReMap.find(integral[j].id);
//     
//     if(mapContent != mastersReMap.end()){
//       
//       integral[j].id = mapContent->second;
//     }
    
    auto findSkipMaster = find(masterVectorSkip.begin(), masterVectorSkip.end(), integral[j].id);
    if ( findSkipMaster != masterVectorSkip.end() ) {
      integral[j].coefficientString = "0";
    }
    
    auto occContent = occurrence.find(integral[j].id);
    if ( occContent != occurrence.end() ) {
      (occContent->second)++;
    }
    else{
      occurrence.insert(pair<std::uint64_t, unsigned > (integral[j].id, 1) );
    }
  }
}


void Kira::load_pyred_on_the_fly(
		std::vector<std::uint64_t>& mandatory, std::vector<std::uint64_t>& optional){
  
  pyred::Config::parse(pyred_config);
  pyred::Config::set_verbosity(2);
  
  vector<pyred::intid> independent_eqnums;
  
  logger << "Load all files from "<<outputDir<<"/tmp/"<<integralfamily.name<<"\n";
  
  ifstream inputConfig((outputDir+"/tmp/"+integralfamily.name+"/SYSTEMconfig").c_str());
  
  numberOfEq = 0;
  while(1){
    
    unsigned numberIND;
    
    if(!(inputConfig >> numberIND)) break;
    
    numberOfEq +=numberIND;
  }
  inputConfig.close();
  
  cout << numberOfEq << endl;
  
  if(skipPyred){
    ofstream output;
    output.open ((outputDir+"/tmp/"+integralfamily.name+"/independentEQS").c_str());
    for(unsigned i = 0; i < numberOfEq; i++) {
      output << i << "\n";
    }
    output.close();
  }
  else{
  
  
    auto sys = pyred::System();
    sys.reserve(numberOfEq);
    
    int eqLength; 
    std::uint64_t ID;
    string line;
    
    std::vector<std::tuple<unsigned,string,int> > equationNumber;
    
    vector<string> add;
    add.push_back("");
    add.push_back("sym");
    
    unsigned countEq = 0;
    for(unsigned k = 0; k < collectReductions.size(); k++) {
      
      for (int sectorNumber = 0; sectorNumber < (1<<integralfamily.jule); 
         sectorNumber++) {
	
        for(int ll = 0; ll<2; ll++){
          
          string inputName = outputDir+"/tmp/" + integralfamily.name + "/SYSTEM"+add[ll]+"_"+collectReductions[k] + "_" + something_string(sectorNumber)+ ".gz";
          
          if(file_exists(inputName.c_str())){  
            igzstream input;
            input.open(inputName.c_str());
            while(1){
              
              vector< pair<std::uint64_t, string > > eq1;
              
              if(!(input >>line)) break;
              if(!(input >>ID)) break;
              if(!(input >> eqLength)) break;
              
              for (int j=0; j<eqLength;j++){
                
                int skip;
                string coef;
                std::uint64_t id;
                
                if(!(input>>skip)) break;
                if(!(input>>coef)) break;
                if(!(input>>id)) break;
                
                eq1.push_back(make_pair(id, coef));
                
                if(!(input>>skip)) break; //sector
                if(!(input>>skip)) break; //topology
                if(!(input>>skip)) break; //flag2
              }
              sys.add_forward(eq1);
	      countEq++;
            }
            input.close();
          }
          load_bar(sectorNumber+1,(1<<integralfamily.jule),50,100);
        }
        
        string inputName = outputDir+"/tmp/" + integralfamily.name + "/SYSTEM"+add[0]+"_"+collectReductions[k] + "_" + something_string(sectorNumber)+ ".gz";
	string inputName2 = outputDir+"/tmp/" + integralfamily.name + "/SYSTEM"+add[1]+"_"+collectReductions[k] + "_" + something_string(sectorNumber)+ ".gz";
        
	if((file_exists(inputName.c_str()) || file_exists(inputName2.c_str()))){
	  
// 	  cout << sectorNumber << " " << (countEq-1) << endl;
	  equationNumber.push_back( std::make_tuple((countEq-1), collectReductions[k], sectorNumber) );
	}
      }
      logger << "\n";
    }
    
    independent_eqnums = sys.backward();
    
    logger << "\nWrite with PYRED_OTF reduced system to hard disk \n";
    logger << "(--> "<<outputDir<<"/tmp/"<<integralfamily.name<<"/independentEQSXXX)\n";
    
    if(mandatory.size()>0){
      
      auto subsys = sys.select(mandatory,optional);
      
      logger << "Reduce: " << subsys.first.size() << " instead of: "
              << independent_eqnums.size() << " integrals\n";
              
      
      record_masters(subsys);
      
      unsigned jEQ = 0;
      unsigned lowerbound = 0;
      ofstream output( (outputDir+"/tmp/"+integralfamily.name+"/independentEQS_" + get<1>(equationNumber[jEQ])+"_"+to_string(get<2>(equationNumber[jEQ]))).c_str() );
	
      for(unsigned i = 0; i < subsys.first.size(); i++) {
      
	while( subsys.first[i] > get<0>(equationNumber[jEQ]) ){
	  
	  lowerbound = get<0>(equationNumber[jEQ])+1;
	  
	  jEQ++;
	  
	  output.close();
	  
	  output.open((outputDir+"/tmp/"+integralfamily.name+"/independentEQS_" + get<1>(equationNumber[jEQ])+"_"+to_string(get<2>(equationNumber[jEQ]))).c_str());
	}
	
	output << (subsys.first[i] - lowerbound) << "\n"; 
	
	if (jEQ > (equationNumber.size()-1))
	  break;
	
      }
      
      output.close();
    }
    else{

      unsigned jEQ = 0;
      unsigned lowerbound = 0;
      ofstream output( (outputDir+"/tmp/"+integralfamily.name+"/independentEQS_" + get<1>(equationNumber[jEQ])+"_"+to_string(get<2>(equationNumber[jEQ]))).c_str() );
      
      for(unsigned i = 0; i < independent_eqnums.size(); i++) {
	
	while( independent_eqnums[i] > get<0>(equationNumber[jEQ]) ){
	  
	  lowerbound = get<0>(equationNumber[jEQ])+1;
	  
	  jEQ++;
	  
	  output.close();
	  
	  output.open((outputDir+"/tmp/"+integralfamily.name+"/independentEQS_" + get<1>(equationNumber[jEQ])+"_"+to_string(get<2>(equationNumber[jEQ]))).c_str());
	}
	
        output << (independent_eqnums[i] - lowerbound) << "\n";
	
	if (jEQ > (equationNumber.size()-1))
	  break;
	
      }
      
      output.close();
    }
    ofstream output( (outputDir+"/tmp/"+integralfamily.name+"/independentEQS"));
    
    if( writeNumericalSystem )
      get_reduced_eqs(sys);
  }  
};

void Kira::get_reduced_eqs(pyred::System & sys){
  
  auto& content = sys.get_content();
  
  ofstream output((outputDir+"/results/"+integralfamily.name+"/reducedEqs").c_str());
  
  for (auto itContent = content.begin(); itContent != content.end(); ++itContent) {
    
    output << "integral ID: " << ((*itContent).first) << " " <<"\n";
    output << "masters ID: ";
    
    auto vecTupel = (std::get<1>((*itContent).second));
    
    for (auto iter = vecTupel.begin(); iter != vecTupel.end(); ++iter) {
      output << *iter << " ";
    }
    output << "\n";
  }
  
  output.close();
};

void Kira::record_masters(std::pair<std::vector<pyred::intid>,std::vector<std::uint64_t> >& subsys){
  
  ofstream output((outputDir+"/results/"+integralfamily.name+"/masters").c_str());
  ofstream tmpoutput((outputDir+"/tmp/"+integralfamily.name+"/masters").c_str());
  
  
  for (unsigned i=0; i<subsys.second.size(); i++){

    auto mapContent = mastersReMap.find(subsys.second[i]);
    
    if(mapContent != mastersReMap.end()){
      subsys.second[i]=mapContent->second;
    }
  }
  
  std::sort(subsys.second.begin(),subsys.second.end());
  subsys.second.resize(distance(subsys.second.begin(),
                                  unique(subsys.second.begin(), subsys.second.end())) );
  
  logger << "Master integrals: (" << subsys.second.size() << ")\n";
  output << "Master integrals: (" << subsys.second.size() << ")\n";
  
  for (unsigned i = 0; i < subsys.second.size(); i++){
    tuple<string,unsigned,unsigned> integral;
    
    get_properties(subsys.second[i],integral);

    logger << subsys.second[i] << ": ";
    logger << collectReductions[get<1>(integral)] << ": ";
    logger << get<0>(integral)<< ": ";
    logger << get<2>(integral);
    logger << "\n";

    output << subsys.second[i] << ": ";
    output << collectReductions[get<1>(integral)] << ": ";
    output << get<0>(integral)<< ": ";
    output << get<2>(integral);
    output << "\n";
    
    tmpoutput << subsys.second[i] << "\n";
  }
  output.close();
}

void Kira::complete_triangular(vector<std::uint64_t>& mandatory){
 
  logger << "\n***** Bring system in triangular form *********************\n";
  
  Clock clock;
  std::map<std::uint64_t, std::tuple<BaseIntegral ***, int> > forwardRed;
  std::vector <std::uint64_t> reduce_please;
  std::unordered_map<std::uint64_t,int> rdy2F;
  
  set<std::uint64_t> setMandatory(mandatory.begin(), mandatory.end());
  
  for(unsigned k = 0; k < collectReductions.size(); k++) {
    
    for (int sectorNumber = 0; sectorNumber < (1<<integralfamily.jule); sectorNumber++) {
      
//       if(file_exists((outputDir+"/tmp/"+integralfamily.name+"/VER"+"_"+
//       collectReductions[k]+"_"+to_string(sectorNumber)+".gz").c_str()))
// 	continue;
      
//       if(file_exists((outputDir+"/tmp/"+integralfamily.name+"/VER"+"_"+
//       collectReductions[k]+"_"+to_string(sectorNumber)+".tmp2.gz").c_str()))
// 	continue;
      
      string inputName = (outputDir+"/tmp/"+integralfamily.name+"/independentEQS_"+collectReductions[k]+"_"+something_string(sectorNumber)).c_str();
    
      if( !(file_exists(inputName.c_str()) ) )
	continue;
      
      load_triangular(k,sectorNumber, forwardRed, reduce_please, rdy2F);
    }
  }
  
  for( FWREDI toIt = forwardRed.begin(); toIt != forwardRed.end(); toIt++){ //collect all keys to iterate
    
    if(get<1>(toIt->second)>1){
      
      reduce_please.push_back(get<0>(toIt->second)[0][0]->id);
    }
  }
  
  run_reduction(forwardRed, reduce_please, rdy2F, setMandatory);
    
  write_triangular(/*sectorNumber, k,*/ forwardRed, 1);
  
  // Clean Up
  for( FWREDI toIt = forwardRed.begin(); toIt != forwardRed.end(); toIt++){ //collect all keys to iterate
    
    if(get<1>(toIt->second)>0){

      delete [] get<0>(toIt->second)[0][0];
    }
    
    delete [] get<0>(toIt->second)[0];
    delete [] get<0>(toIt->second);
  }
  forwardRed.clear();
  rdy2F.clear();
  reduce_please.clear();
  
  logger << "\nTriangular form completed after ( " << clock.eval_time() << " s )\n";
  
  
}

// void Kira::complete_triangularSW(){
//   
//   logger << "\n***** Bring system in triangular form *********************\n";
//   
//   Clock clock;
//   for(unsigned k = 0; k < collectReductions.size(); k++) {
//     
// //     omp_set_nested(1);
// //     omp_set_num_threads(4);
//     #pragma omp parallel for schedule(dynamic)
// //     for (int sectorNumber = (1<<integralfamily.jule)-1; sectorNumber >= 0; sectorNumber--) {
//     for (int sectorNumberI = 0; sectorNumberI < (1<<integralfamily.jule); sectorNumberI++) {
//       int sectorNumber = (1<<integralfamily.jule)-1 - sectorNumberI;
//       
// //       if((sectorNumberI& mask == sectorNumberI))
// // 	cout << "cheers" << endl;
//       
//       std::map<std::uint64_t, std::tuple<BaseIntegral ***, int> > forwardRed;
//       std::vector <std::uint64_t> reduce_please;
//       std::unordered_map<std::uint64_t,int> rdy2F;
//       
//       if(file_exists((outputDir+"/tmp/"+integralfamily.name+"/VER"+"_"+
//       collectReductions[k]+"_"+to_string(sectorNumber)+".gz").c_str()))
// 	continue;
//       
//       if(file_exists((outputDir+"/tmp/"+integralfamily.name+"/VER"+"_"+
//       collectReductions[k]+"_"+to_string(sectorNumber)+".tmp2.gz").c_str()))
// 	continue;
//       
//       string inputName = (outputDir+"/tmp/"+integralfamily.name+"/independentEQS_"+collectReductions[k]+"_"+something_string(sectorNumber)).c_str();
//     
//       if( !(file_exists(inputName.c_str()) ) )
// 	continue;
//       
//       load_triangular(k,sectorNumber, forwardRed, reduce_please, rdy2F);
//       
//       run_reduction(forwardRed, reduce_please, rdy2F);
//     
//       write_triangular(sectorNumber, k, forwardRed,1);
//       
//       // Clean Up
//       for( FWREDI toIt = forwardRed.begin(); toIt != forwardRed.end(); toIt++){ //collect all keys to iterate
// 	
// 	if(get<1>(toIt->second)>0){
// 
// 	  delete [] get<0>(toIt->second)[0][0];
// 	}
// 	
// 	delete [] get<0>(toIt->second)[0];
// 	delete [] get<0>(toIt->second);
//       }
//       forwardRed.clear();
//       rdy2F.clear();
//       reduce_please.clear();
//     }
//     
//     int proc=1;
//     while(proc){
//       proc = 0;
//       for (int sectorNumberI = 0; sectorNumberI < (1<<integralfamily.jule); sectorNumberI++) {
// 	int sectorNumber = (1<<integralfamily.jule)-1 - sectorNumberI;
// 	std::map<std::uint64_t, std::tuple<BaseIntegral ***, int> > forwardRed;
// 	std::vector <std::uint64_t> reduce_please;
// 	std::unordered_map<std::uint64_t,int> rdy2F;
// 	
// 	if(!file_exists((outputDir+"/tmp/"+integralfamily.name+"/SYSTEMx"+"_"+
// 	collectReductions[k]+"_"+to_string(sectorNumber)).c_str())){
// 	    
// 	  rename( (outputDir+"/tmp/"+integralfamily.name+"/VER"+"_"+collectReductions[k]
// 		  +"_"+to_string(sectorNumber)+".tmp2.gz").c_str(), (outputDir+"/tmp/"+integralfamily.name
// 		  +"/VER"+"_"+collectReductions[k]+"_"+to_string(sectorNumber)+".gz").c_str());
// 	  
// 	    continue;
// 	}
// 	
// 	string inputName = (outputDir+"/tmp/"+integralfamily.name+"/independentEQS_"+collectReductions[k]+"_"+something_string(sectorNumber)).c_str();
//       
// 	if( !(file_exists(inputName.c_str()) ) )
// 	  continue;
// 	
// 	load_triangularVER(k,sectorNumber, forwardRed, reduce_please, rdy2F);
// 	
// 	run_reduction(forwardRed, reduce_please, rdy2F);
//       
// 	write_triangular(sectorNumber, k, forwardRed,0);
// 	
// 	proc++;
// 	
// 	// Clean Up
// 	for( FWREDI toIt = forwardRed.begin(); toIt != forwardRed.end(); toIt++){ //collect all keys to iterate
// 	  
// 	  if(get<1>(toIt->second)>0){
// 
// 	    delete [] get<0>(toIt->second)[0][0];
// 	  }
// 	  
// 	  delete [] get<0>(toIt->second)[0];
// 	  delete [] get<0>(toIt->second);
// 	}
// 	forwardRed.clear();
// 	rdy2F.clear();
// 	reduce_please.clear();
//       }
//     }
//   }
//   
//   logger << "\nTriangular form completed after ( " << clock.eval_time() << " s )\n";
//   
// };

int Kira::load_triangular(unsigned k, int sectorNumber, 
			  std::map<std::uint64_t, std::tuple<BaseIntegral ***, int> >& forwardRed, std::vector <std::uint64_t>& reduce_please, 
			  std::unordered_map<std::uint64_t,int>& rdy2F){

  vector<unsigned> independent_eqnums;
  
  ifstream inputConfig;
  string inputNameOTF = (outputDir+"/tmp/"+integralfamily.name+"/independentEQS_"+collectReductions[k]+"_"+something_string(sectorNumber)).c_str();
  
  if(file_exists(inputNameOTF.c_str())){
    inputConfig.open(inputNameOTF.c_str());
//     logger << "read from: " << inputNameOTF << "\n";
    
    while(1){
      
      unsigned numberIND;
      
      if(!(inputConfig >> numberIND)) break;
      
      independent_eqnums.push_back(numberIND);
    }
    inputConfig.close();
  }
//   cout << independent_eqnums << endl;
  
  if(independent_eqnums.size()==0)
    return 0;
  
  unsigned countEq = 0;
  unsigned countALL = 0;
  int breakpoint = 0;
  BaseIntegral *integral = new BaseIntegral [1];
  
  vector<string> add;
  add.push_back("");
  add.push_back("sym");
  
//   for(unsigned k = 0; k < collectReductions.size(); k++) {
    
//     for (int sectorNumber = 0; sectorNumber < (1<<integralfamily.jule); 
//           sectorNumber++) {
      
  for(int ll = 0; ll<2; ll++){
    if(breakpoint)
      break;
    string inputName = outputDir+"/tmp/"+integralfamily.name + "/SYSTEM"+add[ll] + "_"+collectReductions[k] + "_"+something_string(sectorNumber)+".gz";
    if(file_exists(inputName.c_str())){
      
      igzstream input;
      input.open(inputName.c_str());
      while(1){
	
	int eqLength;
	std::uint64_t ID;
	string line;
	if(!(input >> line)) break;
	if(!(input >> ID)) break;
	if(!(input >> eqLength)) break;
	
	if( countALL != independent_eqnums[countEq]){
	  
	  countALL++;
	  
	  for (int j=0; j<eqLength;j++){

	    if(!(input>>integral[0])) break;
	  }
	  continue;
	}
	else{
	  
	  countALL++;
	  countEq++;
	}
	
	read_integral2<igzstream>(input, eqLength, forwardRed, rdy2F);
	if(countEq > (independent_eqnums.size()-1)){
	  breakpoint = 1;
	  break;
	}
      }
      input.close();
    }
//     load_bar(sectorNumber+1,(1<<integralfamily.jule),50,100);
  }
  delete [] integral;
  
  return 1;
};

// int Kira::load_triangularVER(unsigned k, int sectorNumber, 
// 			  std::map<std::uint64_t, std::tuple<BaseIntegral ***, int> >& forwardRed, std::vector <std::uint64_t>& reduce_please, 
// 			  std::unordered_map<std::uint64_t,int>& rdy2F){
//   
//   BaseIntegral *integral = new BaseIntegral [1];
//   
//   string inputName = outputDir+"/tmp/"+integralfamily.name + "/VER" + "_"+collectReductions[k] + "_"+something_string(sectorNumber)+".tmp2.gz";
//   
//   if(file_exists(inputName.c_str())){
//     
//     igzstream input;
//     input.open(inputName.c_str());
//     
//     while(1){
//       
//       int eqLength;
//       std::uint64_t ID;
//       string line;
//       if(!(input >> line)) break;
//       if(!(input >> ID)) break;
//       if(!(input >> eqLength)) break;
//       
//       read_integral2<igzstream>(input, eqLength, forwardRed, rdy2F);
// //       cout << "success1" << endl;
//     }
//     input.close();
//   }
//   
//   inputName = outputDir+"/tmp/"+integralfamily.name + "/SYSTEMx" + "_"+collectReductions[k] + "_"+something_string(sectorNumber);
//   
//   if(file_exists(inputName.c_str())){
//     
//     ifstream input;
//     input.open(inputName.c_str());
//     
//     while(1){
//       
//       int eqLength;
//       std::uint64_t ID;
//       string line;
//       if(!(input >> line)) break;
//       if(!(input >> ID)) break;
//       if(!(input >> eqLength)) break;
//       
//       read_integral2<ifstream>(input, eqLength, forwardRed, rdy2F);
// //       cout << "success2" << endl;
//     }
//     input.close();
//   }
//   
//   for( FWREDI toIt = forwardRed.begin(); toIt != forwardRed.end(); toIt++){ //collect all keys to iterate
//     
//     if(get<1>(toIt->second)>1){
//       get<0>(toIt->second)[0][0]->coefficientString="(-1)";
//       reduce_please.push_back(get<0>(toIt->second)[0][0]->id);
//     }
//   }
//   delete [] integral;
//   
//   return 1;
// };

void Kira::read_line(std::string fileMasters,std::vector<std::pair<std::string,std::vector<int> > >& arraySeed){

  ifstream mastersInput;
  mastersInput.open(fileMasters);
  
  logger << "Look up the seeds.\n";

  string line;
  
  logger.set_level(2);
      
  while(getline(mastersInput, line)) {
    
    //get string indices
    if(line.size()<4)
      continue;
    
    size_t found;
    
    if((found = line.find_first_of("#")) != string::npos){
      logger << "this line will be skept: " << line<< "\n";
      continue;
    }
    
    if((found = line.find_first_of("[")) == string::npos){
      logger << "this line will be skept: " << line<< "\n";
      continue;
    }
    
    line.erase(line.begin(),line.begin()+found+1);
    line.erase(remove_if(line.begin(), line.end(), ::isspace), line.end());
    
    if((found = line.find_first_of("]")) == string::npos){
      logger << "this line will be skept: " << line<< "\n";
      continue;
    }
    
    line.erase(line.begin()+found);//indices
    
//     cout << line << endl;
    
    //convert string indices to integer array
    int indicesCounter;
    vector<int> inputSeed;
    string topo;
    
    if( (found = line.find_first_of(",")) != string::npos){
      istringstream(line.substr(0,found))>>topo;
      line=line.substr(found+1);
    }
    
    while((found = line.find_first_of(",")) != string::npos){
      
      istringstream(line.substr(0,found))>>indicesCounter;
      inputSeed.push_back(indicesCounter);
      line=line.substr(found+1);
      
    }
    
    istringstream(line) >> indicesCounter;
    inputSeed.push_back(indicesCounter);
    arraySeed.push_back(make_pair(topo,inputSeed));
    
  }
  logger.set_level(1);
}

int Kira::load_back_substitution(){
  
  logger << "Load all files from "<<outputDir<<"/tmp/\n";
  
  ifstream inputConfig((outputDir+"/tmp/"+integralfamily.name+"/VERconfig").c_str());
  
  inputConfig >> reihen;
  
  if(reihen == 0){
    logger << "The system you try to reduce contains no equations.\n";
    return 0;
  }
  
  
  DataBase * database = new DataBase(outputDir+"/results/kira.db");
  DataBase * databaseEQ = new DataBase(outputDir+"/results/kira.db");
  
  database[0].select_skipid();
  databaseEQ[0].prepare_lookup_id();
  
  reduct2StartHere.clear();
  occurrence.clear();
  reverseLastReduce.clear();
  
  unsigned skipReihen = 0;
  unsigned reducedReihen = 0;
  
  inputConfig.close();
    
  allEq = new BaseIntegral** [reihen+1];
  last_reduce = new std::uint64_t [reihen+1];
  length = new unsigned [reihen+1];
  rdy2P = new unsigned [reihen+1];
  
  for (unsigned i = 0; i < reihen+1; i++) {
    allEq[i] = new BaseIntegral* [1];
    length[i]=0;
    rdy2P[i]=0;
  }
  
  eqnum=0;
  vector <pair<unsigned,unsigned> > setofS;
  for(unsigned k = 0; k < collectReductions.size(); k++) {
    
    for (int sectorNumber = 0; sectorNumber < (1<<integralfamily.jule); 
         sectorNumber++) {
        
      string inputName = outputDir+"/tmp/"+integralfamily.name+"/VER" + "_"+collectReductions[k] + "_"+something_string(sectorNumber)+".gz";
      
      if(file_exists(inputName.c_str())){
	
	int num_ones=0;
	unsigned testSector=sectorNumber;

	for(size_t i = 0; i < SEEDSIZE; ++i, testSector >>= 1) {
	  if ((testSector & 1) == 1)
	    ++num_ones;
	}
	int it=0;
        igzstream input;
	input.open(inputName.c_str());
	while(1){
	  
	  unsigned eqLength; 
          std::uint64_t ID;
	  string line;
	  if(!(input >> line)) break;
	  if(!(input >> ID)) break;
	  if(!(input >> eqLength)) break;
	  
// 	  if(database[0].find_skipid(ID) && conditionalSystem /*&& conditionalMemory*/){
// 	    
// 	    skip_integral(input, eqLength);
// 	    skipReihen++;
// 	    continue;
// 	  }
	  
	  if(conditionalSystem && databaseEQ[0].bind_id_get_BSequation(ID, allEq[eqnum][0], masterVectorSkip, occurrence)){
	    
	    skip_integral(input,eqLength);
	    reducedReihen++;
	    rdy2P[eqnum]=6;
	  }
	  else{
	    
	    read_integral(allEq[eqnum][0],input,eqLength);
	    rdy2P[eqnum]=3;
	  }
	  
	  last_reduce[eqnum] = ID;
          reverseLastReduce.insert( pair<std::uint64_t, unsigned> (ID, eqnum) );
	  length[eqnum]++;
	  
	  if(num_ones < SEEDSIZE && !(it++)){
	    int flagy = 0;
	    for(size_t itt=0;itt<setofS.size();itt++){
	      if((setofS[itt].first & sectorNumber) == setofS[itt].first && setofS[itt].second == k ){
		flagy = 1;
		break;
	      }
	    }
	    if(!flagy){
	      setofS.push_back(make_pair(sectorNumber,k));
	      reduct2StartHere.push_back(eqnum);
	    }
	  }
	  eqnum++;
	  
	}
	input.close();
      }
    }
  }
  reihen = (reihen - skipReihen);
  logger << "Take " << reihen << " equations to reduce " << (reihen-reducedReihen);
  logger << " missing equations" << "\n";
  
  database[0].finalize();
  delete database;
  databaseEQ[0].finalize();
  delete databaseEQ;
  
  if(reihen == 0){
    logger << "The system is already reduced.\n";
    return 0;
  }
  
  return 1;
  
}

ConvertResult::ConvertResult(Kira& kira, string topologyName_, int topologyNumber_, string &inputName_, const string &outputDir_, vector<std::uint64_t> & idOfSeed){
  
  topologyName = topologyName_;
  topologyNumber = topologyNumber_;
  outputDir = outputDir_;
  inputName = inputName_;
  
  if( !(file_exists(inputName.c_str())) ){
    logger << "\nMissing file with a list of integrals: " << inputName<<"\n";
    logger << "\nUnable to export results to Mathematica\n";
    exit(-1);
  }
  
  seedsInput.open( inputName.c_str() );
  
  look_up_seeds();
  
  database = new DataBase(outputDir+"/results/kira.db"); 
  
  get_integral_id_pyred(idOfSeed);
}

ConvertResult::ConvertResult(Kira& kira, string topologyName_, int topologyNumber_, const string &outputDir_, vector<std::uint64_t> & idOfSeed, string &inputName_){
  
  topologyName = topologyName_;
  topologyNumber = topologyNumber_;
  outputDir = outputDir_;
  inputName = inputName_;

  database = new DataBase(outputDir+"/results/kira.db"); 
}

ConvertResult::~ConvertResult(){
  
 delete database;
 
}

void ConvertResult::get_integral_id_pyred(vector<std::uint64_t> &idOfSeed){  
  
  for(size_t it = 0; it!=arraySeed.size();it++){
    
    auto igl = pyred::Integral(topologyNumber,arraySeed[it]);
    std::uint64_t ID = igl.to_weight();
    if(ID==0){
      logger << "Integral: "<< arraySeed[it] << " is zero.\n";
      continue;
    }
    idOfSeed.push_back(ID); 
  }
}

void ConvertResult::look_up_seeds(){

  logger << "Look up the seeds.\n";

  string line;
  
  logger.set_level(2);
      
  while(getline(seedsInput, line)) {
    
    //get string indices
    if(line.size()<4)
      continue;
    
    size_t found;
    
    if((found = line.find_first_of("#")) != string::npos){
      logger << "this line will be skept: " << line<< "\n";
      continue;
    }
    
    if((found = line.find_first_of("[")) == string::npos){
      logger << "this line will be skept: " << line<< "\n";
      continue;
    }
    
    line.erase(line.begin(),line.begin()+found+1);
    line.erase(remove_if(line.begin(), line.end(), ::isspace), line.end());
    
    if((found = line.find_first_of("]")) == string::npos){
      logger << "this line will be skept: " << line<< "\n";
      continue;
    }
    
    line.erase(line.begin()+found);//indices
    
    //convert string indices to integer array
    int indicesCounter;
    vector<int> inputSeed;
    
    while((found = line.find_first_of(",")) != string::npos){
      
      istringstream(line.substr(0,found))>>indicesCounter;
      inputSeed.push_back(indicesCounter);
      line=line.substr(found+1);
      
    }
    
    istringstream(line) >> indicesCounter;
    inputSeed.push_back(indicesCounter);
    arraySeed.push_back(inputSeed);
    
  }
  logger.set_level(1);
}

void ConvertResult::reconstruct_mass( Kira& kira, vector<DBintegral>& integralV ){

  for(size_t iterator=0;iterator!=integralV.size();iterator++){
    
    int massDimension = integralV[iterator].massDimension-integralV[0].massDimension;

    Perl2Kira perlKira(kira.invar,kira.massSet2One,kira.invarDim,kira.massSet2OneDim);
    perlKira.pipe_kira();
    perlKira.put_pipe(integralV[iterator].coefficient); 
    perlKira.close_pipe();
    perlKira.read_pipe(integralV[iterator].coefficient);
    stringstream ss;
    ss <<"*"<< kira.massSet2One<<"^("<<massDimension*2<<"/"<<kira.massSet2OneDim<<")";
    string MD = ss.str();
    integralV[iterator].coefficient="("+integralV[iterator].coefficient+")"+MD;
    kira.fermat[0].fermat_collect(const_cast<char*>(integralV[iterator].coefficient.c_str()));
    kira.fermat[0].fermat_calc();
    integralV[iterator].coefficient.assign(kira.fermat[0].g_baseout);
  }
}

void ConvertResult::prepare_FORM( vector<DBintegral>& integralV ){
  
  for(size_t iterator=0;iterator!=integralV.size();iterator++){
    
    Perl2Kira perlKira;
    perlKira.pipe_kira();
    perlKira.put_pipe(integralV[iterator].coefficient); 
    perlKira.close_pipe();
    perlKira.read_pipe(integralV[iterator].coefficient);
  }
}

int ConvertResult::output(Kira& kira, int massReconstruction, vector<std::uint64_t>& idOfSeed, int choice){  
  
  kiraOutput outString(choice);
  string reducefile = outputDir+"/results/kira.db";
  
  if( !(file_exists(reducefile.c_str())) ){
    logger << "\nMissing file " << reducefile << ".\n";
    logger << "\nUnable to export results to "<<outString.str[0]<< "\n";
    return 0;
  }
  
  if( !database[0].checkTable()){
    logger << "\nNo reductions found in " << reducefile << ".\n";
    logger << "\nUnable to export results to "<<outString.str[0]<< "\n";
    return 0;
  }
    
  
  string mathematicafile = outputDir+"/results/" +topologyName + "/kira_"+inputName+ outString.str[1];
  Output.open( mathematicafile.c_str() );

  Output << outString.str[2];
    
  for(unsigned int it=0; it!=zeroIntegrals.size(); it++){
    
    Output << outString.str[3] << kira.collectReductions[topologyNumber]<<outString.str[5];
    
    for(unsigned int iterator=0; iterator<zeroIntegrals[it].size();iterator++){
    
      Output <<zeroIntegrals[it][iterator];
     
      if(iterator!=zeroIntegrals[it].size()-1)
	Output<< ",";
    }
    Output << outString.str[7]<<outString.str[4]<<outString.str[10];
    if(it!=zeroIntegrals.size()-1)
      Output << outString.str[6];
  }
  
  database[0].prepare_lookup_id();
  
  int counter=0;
  for(size_t it=0; it!=idOfSeed.size(); it++){
    
    vector<DBintegral> integralV;
    database[0].bind_id_get_equation(idOfSeed[it],integralV);
    
    if(integralV.size()>0)
      counter++;
    
    if(counter==1 && zeroIntegrals.size()>0 && integralV.size()>0)
      Output <<outString.str[6];
    
    if(counter>1 && integralV.size()>0)
      Output <<outString.str[6];
    
    if(massReconstruction)
      reconstruct_mass( kira, integralV );
    
    if(choice == 0)
      prepare_FORM(integralV);
    
    for(size_t iterator=0;iterator!=integralV.size();iterator++){
    
      if(iterator == 0)
	Output << outString.str[3] << kira.collectReductions[integralV[iterator].topology] <<outString.str[5];
      else
	Output << outString.str[4]<<kira.collectReductions[integralV[iterator].topology]<<outString.str[5];
      
      if(choice==2)
	Output << integralV[iterator].id << ",";
      
      Output <<integralV[iterator].indices;
      
      if(iterator==0)
	Output <<outString.str[7];
      else
	Output <<outString.str[8]<<integralV[iterator].coefficient<<outString.str[9];
    }
    
    if(integralV.size()==1)
      Output <<outString.str[10];
    if(integralV.size()==0 && idOfSeed[it] != 0)
      logger << "Integral: " <<idOfSeed[it]<< " is unreduced.\n";
    if(integralV.size()==0 && idOfSeed[it] == 0)
      logger << "One integral is zero.\n";
  }
  
  database[0].finalize();
  
  Output << outString.str[11]<< "\n";
  logger << "\n";
  return 1;
};